/*\
||| Written By Fredrik Hbinette <hubbe@lysator.liu.se>
||| All rights reserved. No warrenties, use at your own risk.
\*/

/*
 * Note to self:
 *
 * Filter functionality for next version:
 *   filter: pnmtogif "$file" "$outfile"
 *   
 * implementation:
 *   o let plugger register a property notification
 *     handler for the plugin window.
 *   o the child will process the data into temporary files
 *     and send the url back to plugger by setting the property
 *   o the child then waits around for a while and deletes
 *     the files afterwards. 
 *   o Make sure that plugger doesn't kill the child, then it
 *     cannot cleanup the temporary files.
 */

#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <sysexits.h>
#include <signal.h>
#include <stdarg.h>
#include <npapi.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/time.h>
#include <X11/X.h>
#include <X11/Xos.h>
#include <X11/Intrinsic.h>
#include <X11/Xatom.h>

#include <sys/socket.h>

#include "plugger.h"

/* UGLY! */
#include "plugger-common.c"

#define WINDOW ((Window) windata.window)

Display *display=0;
NPWindow windata;
int flags;
int repeats;
char *winname;
char *command;
char *file;
int waitfd;

#define FEND(X,Y) (strchr(X,Y)?strchr(X,Y):(X)+strlen(X))
#define xalloc malloc
#define xfree free



/*
 * Functions below are called by the child process
 */

static int childpid=-1;

void kill_child(int tmp)
{
  D("GOT SIGTERM\n");
  if(childpid != -1) my_kill_group(childpid, 0);
  _exit(0);
}


static void very_low_run(char **argv)
{
  setpgrp();
#ifdef H_NOISY
  if(flags & H_NOISY)
  {
    int nl=open("/dev/null", O_RDONLY);
    D("Redirecting stdout and stderr\n");
    dup2(nl,1);
    dup2(nl,2);
    close(nl);
  }
#endif
  
  execvp(argv[0],argv);
  D("Execvp failed..%d\n",errno);
  exit(EX_UNAVAILABLE);
}

int error_handler(Display *dpy, XErrorEvent *err)
{
  D("!!!ERROR_HANDLER!!!\n");
  return 0;
}

#define MAX_IGNORED 1000
static int windows_found=0;
static Window ignored[MAX_IGNORED];

int inital_win_cb(Window w, int depth)
{
  if(windows_found < MAX_IGNORED)
    ignored[windows_found++]=w;
  return 0;
}

int win_cb(Window w, int depth)
{
  int q;
  for(q=0;q<windows_found;q++)
    if(ignored[q]==w)
      return 0;

  return 1;
}


int my_strcmp(char *windowname, char *name)
{
  if(!name) return 1;
  if(!windowname) return 1;

  switch(name[0])
    {
    case '=': return strcmp(name+1, windowname);
    case '~': return strcasecmp(name+1, windowname);
    case '*': return strncasecmp(name+1, windowname, strlen(name)-1);
    }
  return strcasecmp(name, windowname);
}

static Window winrecur(Window from,
		       int depth,
		       char *name,
		       int (*cb)(Window,int))
{
  Window root, parent;
  Window *children=NULL;
  int e,num_children;

  D("Winrecur, checking window %x (depth=%d)\n",from,depth);

  if(!XQueryTree(display, from, &root, &parent, &children, &num_children))
  {
    D("Querytree failed!!!\n");
    return (Window)0;
  }

  D("Num children = %d\n",num_children);

  for (e=0;e<num_children;e++)
  {
    char *windowname;
    unsigned long nitems=0, bytes=0; 
    unsigned char *property=0;
    int fmt;
    Atom type;
    XWindowAttributes a;

    D("XGetWindowAttributes\n");
    if(!XGetWindowAttributes(display, children[e], &a))
      continue;
    if(a.map_state != IsViewable) continue;

    D("XFetchName\n");
    if (0 != XFetchName(display, children[e], &windowname))
    {
      int yes=0;
      D("Winrecur, checking window NAME %x (%s != %s)\n",children[e],windowname,name);
   
      if (!my_strcmp(windowname, name))
	yes=cb(children[e],depth);
      XFree(windowname);
      if(yes) return children[e];
    }

    D("XGetTextProperty\n");
    property=0;
    XGetWindowProperty(display, children[e], XA_WM_CLASS,
		       0, 10000,
		       0,
		       XA_STRING,
		       &type, &fmt, &nitems, &bytes,
		       &property);

    if(property)
    {
      if(type == XA_STRING && fmt==8)
      {
	int len=nitems;
	char *n=(char *)property;
	while(len)
	{
	  D("Winrecur, checking window CLASS %x (%s != %s)\n",children[e],n,name);
	  if (!my_strcmp(n, name))
	    if(cb(children[e],depth))
	      return children[e];
	  
	  len-=strlen(n)+1;
	  n+=strlen(n)+1;
	}
      }
      XFree(property);
    }
  }
  if(depth>0)
  {
    for (e=0;e<num_children;e++)
    {
      Window ret=winrecur(children[e], depth-1, name, cb);
      if(ret) return ret;
    }
  }
  XFree(children);
  return (Window)0;
}


static int gcd(int a, int b)
{
  if (a < b) return gcd(b,a);
  if (b == 0) return a;
  return gcd(b, a % b);
}

static XWindowAttributes wattr;
static int xaspect;
static int yaspect;
static Window victim=0;
static Window old_parent=0;
static Atom lock_property;

static void place_window(void)
{
  int x;
  Window parent=WINDOW;
  int w=windata.width;
  int h=windata.height;

  if(old_parent != parent)
  {
    int q;
    D("Withdrawing!\n");
    XWithdrawWindow(display, victim,
		    XScreenNumberOfScreen (wattr.screen));
    XSync (display, FALSE);

    if(!old_parent)
    {
      XWMHints *leader_change;
      D("Changing leader of window %x\n",victim);
      if((leader_change = XGetWMHints(display, victim)))
      {
	leader_change->flags = (leader_change->flags | WindowGroupHint);
	leader_change->window_group = wattr.root;
	
	XSetWMHints(display,victim,leader_change);
	XFree(leader_change);
      }
    }

    for(x=0;x<25;x++)
    {
      D("Reparenting window %x into %x\n",victim,WINDOW);
      XReparentWindow (display, victim, WINDOW, 0, 0);
      XSync (display, FALSE);
    }
  }else{
    D("Parent is same!\n");
  }

  if(flags & H_FILL)
  {
    D("Resizing window %x with FILL\n",victim);
    D("New size: %dx%d+%d+%d (of %dx%d)\n",w,h,
      0,0,
      w,h);
    
    XMoveResizeWindow(display, victim,
		      0,0,
		      w,h);
  }
  else if(flags & H_MAXASPECT)
  {
    int tmpw, tmph;
    D("Resizing window %x with MAXASPECT\n",victim);
    if(xaspect && yaspect)
    {
      tmph=h/yaspect;
      tmpw=w/xaspect;
      if(tmpw<tmph) tmph=tmpw;
      tmpw=tmph*xaspect;
      tmph=tmph*yaspect;
      
      D("New size: %dx%d+%d+%d (of %dx%d)\n",tmpw,tmph,
	(w - tmpw)/2,
	(h - tmph)/2,
	w,h);
      XMoveResizeWindow(display, victim,
			(w - tmpw)/2,
			(h - tmph)/2,
			tmpw, 
			tmph);
    }
  }

  XMapWindow (display, victim);
  
  XSync (display, FALSE);
  D("placewindow done\n");
}

/* X11 locking mechanism */
static void my_xlockwindow(Display *dpy,
			   Window win,
			   Atom prop)
{
  char buffer[1024];
  struct timeval tv, tv2;

  gettimeofday(&tv, 0);
  gethostname(buffer,sizeof(buffer)-1);
  sprintf(buffer+strlen(buffer),"-%d-%ld:%ld!",
	  getpid(),(long)tv.tv_sec, (long)tv.tv_usec);

  tv.tv_sec+=10; /* Max lock wait time, 10 seconds */
  D("Attempting to lock property = %d!\n",prop);
  D("My LOCKID='%s'\n",buffer);
  while(1)
  {
    int mode=PropModeAppend;
    int fmt;
    XEvent ev;
    unsigned long nitems=0, bytes=0;
    unsigned char *property=0;
    Atom type;

    while(1)
    {
      property=0;
      XGetWindowProperty(dpy, win, prop,
			  0, 1,
			  0, /* Delete */
			  XA_STRING,
			  &type, &fmt, &nitems, &bytes,
			  &property);
      if(!property) break;
      D("Property already set, bytes_left=%d, waiting..\n",bytes);
      XFree(property);
      gettimeofday(&tv2, 0);
      if(tv2.tv_sec > tv.tv_sec)
      {
	D("Stealing lock....\n");
	/* Time to steal the lock! */
	mode=PropModeReplace;
	gettimeofday(&tv, 0);
	INC_TIME(&tv, 0, 10000 + (rand() & 16383));
	break;
      }
      my_usleep((rand() & 16383) + 2000); /* 0.18383 seconds */
    }
    
    D("XChangeProperty, mode=%d\n",mode);
    XChangeProperty(dpy, win, prop, XA_STRING, 8,
		    mode, buffer, strlen(buffer));

    D("Getting property, did we get the lock?\n");
    property=0;
    XGetWindowProperty(dpy, win, prop,
		       0, sizeof(buffer),
		       0, /* Delete */
		       XA_STRING,
		       &type, &fmt, &nitems, &bytes,
		       &property);
    if(property)
    {
      int done=!strncmp(buffer,property,strlen(buffer));
      D("LOCK: done=%d, bytes_left=%d\n",done,bytes);
      XFree(property);
      if(done) return;
    }
  }
}


static void my_xunlockwindow(Display *dpy,
			     Window win,
			     Atom prop)
{
  D("!!!UNLOCKING!!!\n");
  XDeleteProperty(dpy, win, prop);
}



Bool AllXEventsPredicate(Display *dpy, XEvent *ev, char *arg)
{
  return True;
}

int set_aspect(int x, int y)
{
  int tmp;
  int ox=xaspect;
  int oy=yaspect;
  xaspect=x;
  yaspect=y;
  while(1)
  {
    D("computing... xaspect=%d yaspect=%d\n",xaspect,yaspect);
    tmp=gcd(xaspect, yaspect);
    xaspect/=tmp;
    yaspect/=tmp;
    if(xaspect>1 && yaspect>1 && (xaspect+yaspect>100))
    {
      xaspect>>=1;
      yaspect>>=1;
    }else
      break;
  }
  D("xaspect=%d yaspect=%d\n",xaspect,yaspect);
  return ox != xaspect || oy != yaspect;
}

int main(int argc, char **argv)
{
  int do_unlock=0;
  struct timeval lock_timeout;
  struct sigaction action;
  D("Helper started.....\n");

  action.sa_handler=kill_child;
  sigfillset(&action.sa_mask);
  action.sa_flags=0;
  sigaction(SIGTERM, &action, 0);
  sigaction(SIGINT, &action, 0);
  sigaction(SIGTERM, &action, 0);


  if(argc < 2)
  {
    fprintf(stderr,"Plugger version " VERSION " helper application.\n"
	    "Written by Fredrik Hubinette.\n" 
	    "Please see 'man plugger' for details.\n");
    exit(1);
  }

  sscanf(argv[1],"%d,%d,%d,%d,%d,%d,%d,%d",
	 &flags,
	 &repeats,
	 &waitfd,
	 &windata.window,
	 &windata.x,
	 &windata.y,
	 &windata.width,
	 &windata.height);
  command=argv[2];

  winname=getenv("winname");
  file=getenv("file");

  D("HELPER: %s %s %s %s %s\n",
    argv[0],
    argv[1],
    file,
    winname,
    command);

  while(repeats>0)
  {
    char *argv[10];
    char buffer[65536];
    char *foo=buffer;
    int loops=1;

    /* If command expects data on stdin but data is actually in
     * a file, open the file and connect it to stdin.
     /*
    if(file && (flags & H_STREAM))
    {
      int fd=open(file,O_RDONLY);
      dup2(fd,0);
      close(fd);
      D("Stream from file %s\n",file);
    }

    /* Hmm, what does H_REPEATCOUNT do again?? */
    if(flags & H_REPEATCOUNT)
    {
      loops=repeats;
    }

    /* This application will play the data forever */
    if(flags & H_LOOP)
    {
      D("Expecting application to loop.\n");
      loops=0x7fffffff;
    }

    /* setup environment variable $file */
    if(file && !(flags & H_STREAM))
    {
      if((flags & H_MANY) && !(flags & H_REPEATCOUNT))
      {
	int e;
	sprintf(foo,"file=%s",file);
	loops=MIN(repeats,10);
	for(e=0;e<loops;e++)
	{
	  strcat(foo," ");
	  strcat(foo,file);
	}
      }else{
	sprintf(foo,"file=%s",file);
      }
      putenv(foo);
      foo+=strlen(foo)+1;
    }

    /* setup environment variable $window */
    sprintf(foo,"window=%ld",(long)WINDOW);
    putenv(foo);
    foo+=strlen(foo)+1;

    /* setup environment variable $hexwindow */
    sprintf(foo,"hexwindow=0x%lx",(long)WINDOW);
    putenv(foo);
    foo+=strlen(foo)+1;

    /* setup environment variable $repeat */
    sprintf(foo,"repeat=%ld",(long)repeats);
    putenv(foo);
    foo+=strlen(foo)+1;

    sprintf(foo,"xsize=%d",windata.width);
    D("putenv(%s)\n",foo);
    putenv(foo);
    foo+=strlen(foo)+1;
    
    sprintf(foo,"ysize=%d",windata.height);
    D("putenv(%s)\n",foo);
    putenv(foo);
    foo+=strlen(foo)+1;
    
    /* Create command line */
    if(flags & H_CONTROLS)
    {
      argv[0]=getenv("controller");
      argv[1]=command;
      argv[2]=0;
    }else{
      argv[0]="/bin/sh";
      argv[1]="-c";
      argv[2]=command;
      argv[3]=0;
    }

    D("Execing %s (repeats=%d loops=%d)\n",command,repeats,loops);

    if(flags & H_SWALLOW)
    {
      D("Setting up X data (%s)\n",getenv("DISPLAY"));
      if(!display)
      {
	display = XOpenDisplay(getenv("DISPLAY"));
	XSetErrorHandler(error_handler);
      }
      D("display=%x\n",display);
      if(display &&
	 XGetWindowAttributes(display, WINDOW, &wattr))
      {
	victim=0;
	old_parent=0;
	xaspect=yaspect=0;
	D("rootwin=%x\n",wattr.root);
	lock_property=XInternAtom(display, "PLUGGER_SWALLOW_LOCK", 0);
	D("lock_property=%x\n",lock_property);

	my_xlockwindow(display, wattr.root, lock_property);
	do_unlock=1;
	gettimeofday(&lock_timeout, 0);
	lock_timeout.tv_sec+=10; /* ten seconds */

	windows_found=0;
	winrecur(wattr.root, 3, winname, inital_win_cb);

	XSelectInput(display, WINDOW, 
		     SubstructureNotifyMask |
		     SubstructureRedirectMask );

	D("Done Setting up X data\n");
      }else{
	D("XGetWIndowAttributes failed!!!!\n");
	flags &=~ H_SWALLOW;
      }
    }
    
    {
      int pid;
      D("Running %s\n",command);
      pid=fork();
      if(pid==-1) exit(10);
      
      if(!pid)
      {
	very_low_run(argv);
      }else{
	int waitpid_done=0;
	int do_swallow=0;
	int status;
	if(!(flags & H_DAEMON))
	  childpid=pid;
	D("waiting for (%d)\n",pid);

	do_swallow=flags & H_SWALLOW;

	/* Wait for child to die, main event loop */
	while(!waitpid_done)
	{
	  struct timeval tv;
	  fd_set fds;
	  XEvent ev;
	  int maxfd;

	  D("do select\n");

	  FD_ZERO(&fds);
	  FD_SET(waitfd,&fds);
	  maxfd=waitfd;

	  D("do select2\n");
	  D("do sel dpy=%p\n",display);

	  if(display)
	  {
	    FD_SET(ConnectionNumber(display),&fds);
	    maxfd=MAXIMUM(maxfd, ConnectionNumber(display));
	  }

	  D("do select3\n");

	  tv.tv_sec=0;

	  D("do select4\n");
	  if(do_swallow && do_unlock)
	  {
	    tv.tv_usec=1000000/50;
	  }else{
	    tv.tv_usec=1000000/5;
	  }
	  D("do select5\n");
	  select( maxfd + 1, &fds, NULL, NULL, &tv);

	  D("select done\n");
	  
	  while(display &&
		XCheckIfEvent(display, &ev, AllXEventsPredicate, NULL))
	  {
	    D("got event\n");
	    switch(ev.type)
	    {
	      case DestroyNotify:
		if(flags & H_SWALLOW) do_swallow=1;
		break;

	      case ConfigureRequest:
		if(set_aspect(ev.xconfigure.width, ev.xconfigure.height))
		  place_window();
		break;

	      default:
		D("Received unknown event: %d\n",ev.type);
		break;
	    }
	  }

	  if(do_swallow)
	  {
	    /* Find window and reparent it */
	    int tmp;
	    D("Looking for victim... (%s)\n",winname);
	    if((victim=winrecur(wattr.root, 3, winname, win_cb)))
	    {
	      D("Found it!! Victim=%x\n",victim);
	      if(do_unlock)
	      {
		my_xunlockwindow(display, wattr.root, lock_property);
		do_unlock=0;
	      }
	      
	      if(flags & H_MAXASPECT)
	      {
		int tmp;
		XWindowAttributes ca;
		D("MAXASPECT\n");
		XGetWindowAttributes(display, victim, &ca);
		set_aspect(ca.width,ca.height);
	      }
	      
	      place_window();
	      do_swallow=0;
	    }
	  }

	  if(do_unlock)
	  {
	    struct timeval tv;
	    gettimeofday(&tv, 0);
	    if(tv.tv_sec > lock_timeout.tv_sec)
	    {
	      my_xunlockwindow(display, wattr.root, lock_property);
	      do_unlock=0;
	    }
	  }
	  

	  if(FD_ISSET(waitfd, &fds))
	  {
	    /* Re-reparent window */
	    static NPWindow wintmp;
	    static int rptr=0;
	    int tmp;

	    Window oldwindow=WINDOW;
	    D("Got waitfd data, old parent=%x waitfd=%d\n",WINDOW,waitfd);
	    
	    tmp=read(waitfd, ((char *)& wintmp)+ rptr,  sizeof(wintmp) - rptr);
	    if(tmp < 0)
	    {
	      if(errno == EINTR) goto incomplete_windata;
	      D("Winddata read error, exiting\n");
	      _exit(1);
	    }

	    if(tmp == 0)
	    {
	      D("Winddata EOF, exiting\n");
	      _exit(1);
	    }
	    rptr+=tmp;

	    if(rptr != sizeof(windata)) goto incomplete_windata;
	    
	    windata = wintmp;
	    rptr=0;

	    D("Got waitfd data, new parent=%x\n",WINDOW);

	    XSelectInput(display, WINDOW, 
			 SubstructureNotifyMask |
			 SubstructureRedirectMask );

	    if(victim)
	    {
	      place_window();
	    }else{
	      if((my_strstr(command,"$window") ||
		  my_strstr(command,"$hexwindow")) &&
		 oldwindow != WINDOW)
	      {
		/*
		 * Kill and re-start child
		 * Hmm, what happens if we are streaming????
		 * We're not (because Mozilla can't)
		 */
		my_kill_group(pid,0);
		waitpid_done++;
		continue;
	      }
	    }

	  incomplete_windata: ;
	  }

	  D("Waiting for godot... %d\n",pid);

	  if(!((flags & H_EXITS) || waitpid_done))
	  {
	    if(waitpid(pid,&status,WNOHANG) > 0)
	      waitpid_done++;
	  }
	  D("Wait done?? (%d)\n",waitpid_done);
	}

	if(do_unlock)
	{
	  my_xunlockwindow(display, wattr.root, lock_property);
	  do_unlock=0;
	}

	D("wait done repeats=%d loops=%d\n",repeats, loops);
	if(!WIFEXITED(status))
	{
	  D("Process dumped core or something...\n");
	  exit(10);
	}
	if(WEXITSTATUS(status) && !(flags & H_IGNORE_ERRORS))
	{
	  D("Process exited with error code: %d\n",WEXITSTATUS(status));
	  exit(WEXITSTATUS(status));
	}
	D("exited ok!\n");
      }
      if(repeats < MAXINT)
	repeats-=loops;
    }
  }
  D("EXITING\n");
  _exit(0);
}
